"""Tests for the v2 rest api"""
import json
import pytest
import requests_mock as rmock
from bitfinex.rest import ClientV2 as Client

# pylint: disable=W0621,C0111

@pytest.fixture
def client():
    return Client("key", "secret")


def test_platform_status(client, requests_mock):
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text='[1]'
    )
    platform_status = client.platform_status()
    assert platform_status == [1]


def test_tickers_url_is_ok(client, requests_mock):
    response_text = "[]"
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.tickers(['tIOTBTC', 'tIOTUSD'])
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/tickers?symbols=tIOTBTC,tIOTUSD'
    )

def test_tickers_response(client, requests_mock):
    response_text = json.dumps([
        ['tIOTBTC',
         0.00012012,
         207246.56609548,
         0.00012048,
         230371.29034138,
         -8.28e-06,
         -0.0644,
         0.00012035,
         3215468.26026294,
         0.00012863,
         0.00011545],
        ['tIOTUSD',
         0.95969,
         24306.55703175,
         0.96214,
         60186.47595187,
         -0.03534,
         -0.0355,
         0.95925,
         13274504.88034301,
         0.99459, 0.90661]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    tickers_response = client.tickers(['tIOTBTC', 'tIOTUSD'])
    assert isinstance(tickers_response, list)
    assert tickers_response[0][0] == "tIOTBTC"


def test_ticker_url_is_ok(client, requests_mock):
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text="[]"
    )
    client.ticker('tIOTBTC')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/ticker/tIOTBTC'
    )

def test_ticker_response(client, requests_mock):
    response_text = json.dumps([
        0.00011989,
        224267.1563644,
        0.00012015,
        265388.46405117,
        -8.12e-06,
        -0.0634,
        0.00012,
        3206446.72556849,
        0.00012812,
        0.00011545
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ticker_response = client.ticker('tIOTBTC')
    assert isinstance(ticker_response, (list,))




def test_trades_url_is_ok(client, requests_mock):
    response_text = "[]"
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.trades('tIOTUSD')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/trades/tIOTUSD/hist'
    )

def test_trades_response(client, requests_mock):
    response_text = json.dumps([
        [272147102, 1532428225632, -10.93936053, 0.00012022],
        [272147036, 1532428201910, -260.98846449, 0.00012028]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    trades_response = client.trades('tIOTUSD')
    assert isinstance(trades_response, list)
    assert trades_response[0][0] == 272147102

def test_books_url_is_ok(client, requests_mock):
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text="[]"
    )
    client.books('tIOTUSD', 'P1')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/book/tIOTUSD/P1'
    )

def test_books_response(client, requests_mock):
    response_text = json.dumps([
        [0.97692, 2, 1476.41706974],
        [0.97446, 1, 162.69181129],
        [0.97428, 3, 471.48294762]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    books_response = client.books('tIOTUSD', 'P1')
    assert isinstance(books_response, list)
    assert books_response[0][0] == 0.97692

def test_stats_url_is_ok(client, requests_mock):
    response_text = "[]"
    params = {
        'key'     : 'funding.size',
        'size'    : '1m',
        'symbol'  : 'fUSD',
        'section' : 'hist',
        'sort'    : '0'
    }
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.stats(**params)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/stats1/funding.size:1m:fUSD/hist?sort=0'
    )

def test_stats_response(client, requests_mock):
    response_text = json.dumps([
        [1532430780000, 548711270.8708785],
        [1532430720000, 546414815.6475447],
        [1532430660000, 545031610.6767836]
    ])
    params = {
        'key'     : 'funding.size',
        'size'    : '1m',
        'symbol'  : 'fUSD',
        'section' : 'hist',
        'sort'    : '0'
    }
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    stats_response = client.stats(**params)
    assert isinstance(stats_response, list)
    assert stats_response[0][0] == 1532430780000
    assert stats_response[0][1] == 548711270.8708785

def test_candles_url_is_ok(client, requests_mock):
    response_text = json.dumps([
        [1532431320000, 8140.8, 8138.1, 8144, 8138, 15.50046363],
        [1532431260000, 8140.1, 8139.9, 8143.9, 8138.00802545, 29.16099656]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.candles("1m", "tBTCUSD", "hist", limit='1')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/candles/trade:1m:tBTCUSD/hist?limit=1'
    )

def test_candles_response(client, requests_mock):
    response_text = json.dumps([
        [1532431320000, 8140.8, 8138.1, 8144, 8138, 15.50046363],
        [1532431260000, 8140.1, 8139.9, 8143.9, 8138.00802545, 29.16099656]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    candles_response = client.candles("1m", "tBTCUSD", "hist", limit='1')
    assert isinstance(candles_response, list)
    assert candles_response[0][0] == 1532431320000
    assert candles_response[0][1] == 8140.8


def test_market_avg_url_is_ok(client, requests_mock):
    response_text = json.dumps([7912.26508244, 100])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.market_average_price(symbol="tBTCUSD", amount="100", period="1m")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/calc/trade/avg'
    )

def test_market_avg_response(client, requests_mock):
    response_text = json.dumps([7912.26508244, 100])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    map_response = client.market_average_price(symbol="tBTCUSD", amount="100", period="1m")
    assert isinstance(map_response, list)
    assert map_response[0] == 7912.26508244
    assert map_response[1] == 100



def test_fx_rate_url_is_ok(client, requests_mock):
    response_text = json.dumps([1.00105])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.foreign_exchange_rate(ccy1="IOT", ccy2="USD")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/calc/fx'
    )


def test_fx_rate_response(client, requests_mock):
    response_text = json.dumps([1.00105])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    map_response = client.foreign_exchange_rate(ccy1="IOT", ccy2="USD")
    assert isinstance(map_response, list)
    assert map_response[0] == 1.00105


def test_wallets_balance_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.wallets_balance()
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/wallets'
    )

def test_wallets_balance_response(client, requests_mock):
    response_text = json.dumps([
        ['funding', 'IOT', 6380, 0, None],
        ['funding', 'OMG', 7.74457581, 0, None],
        ['exchange', 'BAT', 5031.10567614, 0, None],
        ['exchange', 'IOT', 3957.20410603, 0, None],
        ['exchange', 'MNA', 19.05314345, 0, None],
        ['exchange', 'USD', 0.62804524, 0, None],
        ['margin', 'BTC', 0, 0, None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    wb_response = client.wallets_balance()
    assert isinstance(wb_response, list)
    assert wb_response[0][0] == 'funding'
    assert wb_response[0][1] == 'IOT'
    assert wb_response[0][2] == 6380


def test_active_orders_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.active_orders("tIOTUSD")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/orders/tIOTUSD'
    )

def test_active_orders_response(client, requests_mock):
    response_text = json.dumps([
        [13919467011,
         None,
         40580406675,
         'tIOTUSD',
         1530530180426,
         1530530180438,
         -4006.8,
         -4006.8,
         'EXCHANGE LIMIT',
         None,
         None,
         None,
         0,
         'ACTIVE',
         None,
         None,
         8.08,
         0,
         None,
         None,
         None,
         None,
         None,
         0,
         0,
         0,
         None,
         None,
         'API>BFX',
         None,
         None,
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ao_response = client.active_orders("tIOTUSD")
    assert isinstance(ao_response, list)
    assert ao_response[0][3] == 'tIOTUSD'


def test_orders_history_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.orders_history("tIOTUSD")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/orders/tIOTUSD/hist'
    )

def test_orders_history_response(client, requests_mock):
    response_text = json.dumps([
        [14756198784,
         '71',
         688,
         'tIOTUSD',
         1532587404000,
         1532675646000,
         0,
         50.05,
         'EXCHANGE LIMIT',
         None,
         None,
         None,
         None,
         'EXECUTED @ 1.0(50.05)',
         None,
         None,
         1,
         1,
         0,
         0,
         None,
         None,
         None,
         0,
         0,
         None,
         None,
         None,
         'API>BFX',
         None,
         None,
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    oh_response = client.orders_history("tIOTUSD")
    assert isinstance(oh_response, list)
    assert oh_response[0][3] == 'tIOTUSD'


def test_order_trades_url_is_ok(client, requests_mock):
    response_text = json.dumps([
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.order_trades("tIOTUSD", 14395751815)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/order/tIOTUSD:14395751815/trades'
    )

def test_order_trades_response(client, requests_mock):
    response_text = json.dumps([
        [269943238,
         'tIOTUSD',
         1531926144000,
         14463187162,
         50.05,
         1.1272,
         None,
         None,
         1,
         -0.05641636,
         'USD']
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ot_response = client.order_trades("tIOTUSD", 14395751815)
    assert isinstance(ot_response, list)
    assert ot_response[0][1] == 'tIOTUSD'


def test_trades_history_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.trades_history('tIOTUSD', limit=10)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/trades/tIOTUSD/hist'
    )

def test_trades_history_response(client, requests_mock):
    response_text = json.dumps([
        [273186883,
         'tIOTUSD',
         1532675647000,
         14756198784,
         50.05,
         1,
         None,
         None,
         None,
         -0.05005,
         'USD'],
        [273177908,
         'tIOTUSD',
         1532674331000,
         14756270025,
         50.05,
         1.005,
         None,
         None,
         None,
         -0.05030025,
         'USD']
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    th_response = client.trades_history('tIOTUSD', limit=10)
    assert isinstance(th_response, list)
    assert th_response[0][1] == 'tIOTUSD'
    assert th_response[1][1] == 'tIOTUSD'


def test_active_positions_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.active_positions()
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/positions'
    )

def test_active_positions_response(client, requests_mock):
    response_text = json.dumps([
        ['tIOTUSD',
         'ACTIVE',
         4000,
         1.02400594,
         -0.12578296,
         0,
         -193.44454356,
         -4.7227398,
         0.3649879,
         0.8713]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ap_response = client.active_positions()
    assert isinstance(ap_response, list)
    assert ap_response[0][0] == 'tIOTUSD'
    assert ap_response[0][1] == 'ACTIVE'


def test_funding_offers_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_offers("fIOT")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/offers/fIOT'
    )

def test_funding_offers_response(client, requests_mock):
    response_text = json.dumps([
        [277750009,
         'fIOT',
         1532669718000,
         1532669718000,
         30000,
         30000,
         'LIMIT',
         None,
         None,
         0,
         'ACTIVE',
         None,
         None,
         None,
         0.002,
         10,
         0,
         0,
         None,
         0,
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    fo_response = client.funding_offers("fIOT")
    assert isinstance(fo_response, list)
    assert fo_response[0][1] == 'fIOT'
    assert fo_response[0][10] == 'ACTIVE'




def test_funding_offer_history_url(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_offers_history('fOMG')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/offers/fOMG/hist'
    )

def test_funding_offers_hist_resp(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    foh_response = client.funding_offers_history('fOMG')
    assert isinstance(foh_response, list)




def test_funding_loans_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_loans("fIOT")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/loans/fIOT'
    )

def test_funding_loans_response(client, requests_mock):
    response_text = json.dumps([
        [7708021,
         'fOMG',
         1,
         1530761089000,
         1532693100000,
         7.74107199,
         0,
         'ACTIVE',
         None,
         None,
         None,
         0,
         30,
         1530761089000,
         1532693100000,
         0,
         0,
         None,
         0,
         None,
         0]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    fl_response = client.funding_loans("fIOT")
    assert isinstance(fl_response, list)
    assert fl_response[0][1] == 'fOMG'
    assert fl_response[0][7] == 'ACTIVE'



def test_funding_loans_history_url(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_loans_history('fOMG')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/loans/fOMG/hist'
    )

def test_funding_loans_history_resp(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    flh_response = client.funding_loans_history('fOMG')
    assert isinstance(flh_response, list)



def test_funding_credits_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_credits("fIOT")
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/credits/fIOT'
    )

def test_funding_credits_response(client, requests_mock):
    response_text = json.dumps([
        [104425097,
         'fUSD',
         -1,
         1532648567000,
         1532696761000,
         1011.1101,
         0,
         'ACTIVE',
         None,
         None,
         None,
         0.00013245,
         2, 1532648567000,
         1532696761000,
         0,
         0,
         None,
         0,
         None,
         0,
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    fc_response = client.funding_credits("fUSD")
    assert isinstance(fc_response, list)
    assert fc_response[0][1] == 'fUSD'
    assert fc_response[0][7] == 'ACTIVE'


def test_funding_credits_history_url(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_credits_history('fUSD')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/credits/fUSD/hist'
    )

def test_funding_credits_history_resp(client, requests_mock):
    response_text = json.dumps([
        [104187074,
         'fUSD',
         1,
         1532539366000,
         1532589764000,
         518.749995,
         None,
         'CLOSED (reduce_swaps)',
         None,
         None,
         None,
         0.0001109,
         2,
         None,
         None,
         0,
         0,
         None,
         0,
         None,
         0,
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    fch_response = client.funding_credits_history('fUSD')
    assert isinstance(fch_response, list)
    assert fch_response[0][1] == 'fUSD'


def test_funding_trades_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_trades('fUSD')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/funding/trades/fUSD/hist'
    )

def test_funding_trades_response(client, requests_mock):
    response_text = json.dumps([
        [71126559, 'fUSD', 1532690564000, 0, -489.93446, 0.00013558, 2, None],
        [71124554, 'fUSD', 1532689747000, 0, -491.490995, 0.00011935, 2, None],
        [71114966, 'fUSD', 1532683364000, 0, -498.1977, 0.00013667, 2, None],
        [71100173, 'fUSD', 1532676165000, 0, -500.70021, 0.00013111, 2, None],
        [71047501, 'fUSD', 1532653969000, 0, -0.28881967, 0.00014703, 2, None],
        [71034152, 'fUSD', 1532648567000, 0, -1011.1101, 0.00013245, 2, None],
        [70985009, 'fUSD', 1532628548000, 0, -912.55269924, 0.0001188, 2, None],
        [70985008, 'fUSD', 1532628548000, 0, -67.44734076, 0.00011779, 2, None],
        [70869292, 'fUSD', 1532569366000, 0, -0.54663117, 0.0001355, 2, None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ft_response = client.funding_trades('fUSD')
    assert isinstance(ft_response, list)
    assert ft_response[0][1] == 'fUSD'

def test_margin_info_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.margin_info('tIOTUSD')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/info/margin/tIOTUSD'
    )

def test_margin_info_response(client, requests_mock):
    response_text = json.dumps([
        'sym',
        'tIOTUSD',
        [9018.05035061, 14876.97279239, 9018.05035061, 7894.15035061, None, None, None, None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    mi_response = client.margin_info('tIOTUSD')
    assert isinstance(mi_response, list)
    assert mi_response[1] == 'tIOTUSD'
    assert isinstance(mi_response[2], list)


def test_funding_info_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.funding_info('fUSD')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/info/funding/fUSD'
    )

def test_funding_info_response(client, requests_mock):
    response_text = json.dumps([
        'sym',
        'fUSD',
        [0, 0, 1.52802813, 0]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    fi_response = client.funding_info('fUSD')
    assert isinstance(fi_response, list)
    assert fi_response[0] == 'sym'
    assert fi_response[1] == 'fUSD'
    assert isinstance(fi_response[2], list)



def test_movements_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.movements('IOT')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/movements/IOT/hist'
    )

def test_movements_response(client, requests_mock):
    response_text = json.dumps([
        [2550180,
         'IOT',
         'IOTA',
         None,
         None,
         1500445999000,
         1500445999000,
         None,
         None,
         'COMPLETED',
         None,
         None,
         1800,
         0,
         None,
         None,
         'UWDERLMBXDMVVYYLPCGLORYTOTHEROMANEMPIRERBRYANODCLYAFMGFCRULMENTULBARLADSHZDQHCMOMNUVJXGDRS',
         None,
         None,
         None,
         '2550035',
         None]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    m_response = client.movements('IOT')
    assert isinstance(m_response, list)
    assert m_response[0][1] == 'IOT'
    assert m_response[0][9] == 'COMPLETED'


def test_performance_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.performance()
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/stats/perf:1D/hist'
    )


def test_alert_list_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.alert_list()
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/alerts'
    )

def test_alert_list_response(client, requests_mock):
    response_text = json.dumps([
        ['price:tEOSUSD:14.524', 'price', 'tEOSUSD', 14.524, 93],
        ['price:tBTCUSD:6264.3', 'price', 'tBTCUSD', 6264.3, 100],
        ['price:tBCHUSD:713.39', 'price', 'tBCHUSD', 713.39, 100],
        ['price:tBTCUSD:6380.8', 'price', 'tBTCUSD', 6380.8, 94],
        ['price:tBCHUSD:770.08', 'price', 'tBCHUSD', 770.08, 27],
        ['price:tIOTUSD:1.0817', 'price', 'tIOTUSD', 1.0817, 100],
        ['price:tIOTUSD:0.97758', 'price', 'tIOTUSD', 0.97758, 74],
        ['price:tIOTUSD:0.99984', 'price', 'tIOTUSD', 0.99984, 100]
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    al_response = client.alert_list()
    assert isinstance(al_response, list)
    assert al_response[0][2] == 'tEOSUSD'


def test_alert_set_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.alert_set('price', 'tIOTUSD', 1)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/w/alert/set'
    )

def test_alert_set_response(client, requests_mock):
    response_text = json.dumps([
        'price:tIOTUSD:3',
        'price',
        'tIOTUSD',
        3,
        100
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    as_response = client.alert_set('price', 'tIOTUSD', 1)
    assert isinstance(as_response, list)
    assert as_response[2] == 'tIOTUSD'
    assert as_response[3] == 3



def test_alert_delete_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.alert_delete('tIOTUSD', 3)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/w/alert/price:tIOTUSD:3/del'
    )

def test_alert_delete_response(client, requests_mock):
    response_text = json.dumps([True])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    ad_response = client.alert_delete('tIOTUSD', 3)
    assert isinstance(ad_response, list)
    assert ad_response[0]



def test_calc_available_balance_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.calc_available_balance('tIOTUSD', 1, 1.13, 'EXCHANGE')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/calc/order/avail'
    )

def test_calc_available_balance_response(client, requests_mock):
    response_text = json.dumps([0.79734514])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    wb_response = client.calc_available_balance('tIOTUSD', 1, 1.13, 'EXCHANGE')
    assert isinstance(wb_response, list)
    assert wb_response[0] == 0.79734514


def test_ledgers_url_is_ok(client, requests_mock):
    response_text = json.dumps([])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    client.ledgers('IOT')
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v2/auth/r/ledgers/IOT/hist'
    )

def test_ledger_response(client, requests_mock):
    response_text = json.dumps([
        [1772673563,
         'IOT',
         None,
         1532115843000,
         None,
         40,
         2672.7521789,
         None,
         'Transfer of 40.0 IOT from wallet Exchange to Trading on wallet margin'],
        [1772673562,
         'IOT',
         None,
         1532115843000,
         None,
         -40,
         0.81645596,
         None,
         'Transfer of 40.0 IOT from wallet Exchange to Trading on wallet exchange']
    ])
    requests_mock.register_uri(
        rmock.ANY,
        rmock.ANY,
        text=response_text
    )
    l_response = client.ledgers('IOT')
    assert isinstance(l_response, list)
    assert l_response[0][1] == 'IOT'
    assert l_response[1][1] == 'IOT'



# coding=utf-8
import os
import threading
from datetime import date
import json
import hmac, hashlib

from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol, \
    connectWS
from twisted.internet import reactor, ssl
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.error import ReactorAlreadyRunning
from bitfinex.websockets import wss_utils

# Example used to make send logic
# https://stackoverflow.com/questions/18899515/writing-an-interactive-client-with-twisted-autobahn-websockets

class BitfinexClientProtocol(WebSocketClientProtocol):

    def __init__(self, factory, payload=None):
        super().__init__()
        self.factory = factory
        self.payload = payload

    def onOpen(self):
        self.factory.protocol_instance = self

    def onConnect(self, response):
        if self.payload:
            self.sendMessage(self.payload, isBinary=False)
        # reset the delay after reconnecting
        self.factory.resetDelay()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            try:
                payload_obj = json.loads(payload.decode('utf8'))
            except ValueError:
                pass
            else:
                self.factory.callback(payload_obj)

class BitfinexReconnectingClientFactory(ReconnectingClientFactory):

    # set initial delay to a short time
    initialDelay = 0.1

    maxDelay = 20

    maxRetries = 30


class BitfinexClientFactory(WebSocketClientFactory, BitfinexReconnectingClientFactory):

    def __init__(self, *args, payload=None, **kwargs):
        WebSocketClientFactory.__init__(self, *args, **kwargs)
        self.protocol_instance = None
        self.base_client = None
        self.payload = payload

    protocol = BitfinexClientProtocol
    _reconnect_error_payload = {
        'e': 'error',
        'm': 'Max reconnect retries reached'
    }

    def clientConnectionFailed(self, connector, reason):
        self.retry(connector)
        if self.retries > self.maxRetries:
            self.callback(self._reconnect_error_payload)

    def clientConnectionLost(self, connector, reason):
        self.retry(connector)
        if self.retries > self.maxRetries:
            self.callback(self._reconnect_error_payload)

    def buildProtocol(self, addr):
        return BitfinexClientProtocol(self, payload=self.payload)


class BitfinexSocketManager(threading.Thread):

    STREAM_URL = 'wss://api.bitfinex.com/ws/2'

    def __init__(self): #client
        """Initialise the BitfinexSocketManager"""
        threading.Thread.__init__(self)
        self.factories = {}
        self._connected_event = threading.Event()
        self._conns = {}
        self._user_timer = None
        self._user_listen_key = None
        self._user_callback = None

    def _start_socket(self, id_, payload, callback):
        if id_ in self._conns:
            return False

        factory_url = self.STREAM_URL
        factory = BitfinexClientFactory(factory_url, payload=payload)
        factory.base_client = self
        factory.protocol = BitfinexClientProtocol
        factory.callback = callback
        factory.reconnect = True
        context_factory = ssl.ClientContextFactory()
        self.factories[id_] = factory
        reactor.callFromThread(self.add_connection, id_)


    def add_connection(self, id_):
        """
        Convenience function to connect and store the resulting
        connector.
        """
        factory = self.factories[id_]
        context_factory = ssl.ClientContextFactory()
        self._conns[id_] = connectWS(factory, context_factory)


    def stop_socket(self, conn_key):
        """Stop a websocket given the connection key
        :param conn_key: Socket connection key
        :type conn_key: string
        :returns: connection key string if successful, False otherwise
        """
        if conn_key not in self._conns:
            return

        # disable reconnecting if we are closing
        self._conns[conn_key].factory = WebSocketClientFactory(self.STREAM_URL)
        self._conns[conn_key].disconnect()
        del(self._conns[conn_key])


    def run(self):
        try:
            reactor.run(installSignalHandlers=False)
        except ReactorAlreadyRunning:
            # Ignore error about reactor already running
            pass

    def close(self):
        """Close all connections
        """
        keys = set(self._conns.keys())
        for key in keys:
            self.stop_socket(key)

        self._conns = {}


class WssClient(BitfinexSocketManager):

    ###########################################################################
    # Bitfinex commands
    ###########################################################################

    def __init__(self, key=None, secret=None): #client
        """Initialise the WssClient"""
        super().__init__()
        self.key = key
        self.secret = secret

    def authenticate(self, callback, filters=None):
        nonce = int(wss_utils.UtcNow() * 1000000)
        auth_payload = 'AUTH{}'.format(nonce)
        signature = hmac.new(
            self.secret.encode(), #settings.API_SECRET.encode()
            msg=auth_payload.encode('utf8'),
            digestmod=hashlib.sha384
        ).hexdigest()
        data = {
            # docs: http://bit.ly/2CEx9bM
            'event': 'auth',
            'apiKey': self.key,
            'authSig': signature,
            'authPayload': auth_payload,
            'authNonce': nonce,
            'calc': 1
        }
        if filters:
            data['filter'] = filters
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket("auth", payload, callback)

    def subscribe_to_ticker(self, symbol, callback):
        """Subscribe to the passed symbol ticks data channel.
        """
        symbol = 't' + symbol if not symbol.startswith('t') else symbol
        id_ = "_".join(["ticker", symbol])
        data = {
            'event': 'subscribe',
            'channel': 'ticker',
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    def subscribe_to_trades(self, symbol, callback):
        """Subscribe to the passed symbol trades data channel.
        """
        symbol = 't' + symbol if not symbol.startswith('t') else symbol
        id_ = "_".join(["trades", symbol])
        data = {
            'event': 'subscribe',
            'channel': 'trades',
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    #Precision: R0, P0, P1, P2, P3
    def subscribe_to_orderbook(self, symbol, precision, callback):
        """Subscribe to the orderbook of a given pair.

        Parameters
        ----------
        symbol : str
            Symbol pair to request data for.

        precision : str
            Accepted values as strings {R0, P0, P1, P2, P3}

        callback : func
            A function to use as callback
        """
        symbol = 't' + symbol if not symbol.startswith('t') else symbol
        id_ = "_".join(["order", symbol])
        data = {
            'event': 'subscribe',
            "channel": "book",
            "prec": precision,
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    # TODO: Switch "pair" to "symbol" in new major version
    def subscribe_to_candles(self, pair, timeframe, callback):
        """Subscribe to the passed pair's OHLC data channel.

        Parameters
        ----------
        pair : str
            Symbol pair to request data for

        timeframe : str
            Accepted values as strings {1m, 5m, 15m, 30m, 1h, 3h, 6h, 12h,
            1D, 7D, 14D, 1M}

        callback : func
            A function to use as callback

        Returns
        -------
        str
            The socket identifier.

        """

        valid_tfs = ['1m', '5m', '15m', '30m', '1h', '3h', '6h', '12h', '1D',
                     '7D', '14D', '1M']
        if timeframe:
            if timeframe not in valid_tfs:
                raise ValueError("timeframe must be any of %s" % valid_tfs)
        else:
            timeframe = '1m'
        identifier = ('candles', pair, timeframe)
        id_ = "_".join(identifier)
        pair = 't' + pair if not pair.startswith('t') else pair
        key = 'trade:' + timeframe + ':' + pair
        data = {
            'event': 'subscribe',
            'channel': 'candles',
            'key': key,
        }
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    def ping(self, channel="auth"):
        """Ping bitfinex."""
        client_cid = wss_utils.UtcNow()
        data = {
            'event': 'ping',
            'cid': client_cid
        }
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        self.factories[channel].protocol_instance.sendMessage(payload, isBinary=False)
        return client_cid

    def new_order_op(self, order_type, pair, amount, price, price_trailing=None,
                     price_aux_limit=None, price_oco_stop=None, hidden=0,
                     flags=None, tif=None):
        """
        Create new order operation

        Parameters
        ----------
        order_type : str
            Order type. Must be one of: "LIMIT", "STOP", "MARKET",
            "TRAILING STOP", "FOK", "STOP LIMIT" or equivelent with "MARKET"
            prepended to it.

        pair : str
            The currency pair to be traded. e.g. BTCUSD

        amount :  float
            The amount to be traided.

        price : float
            The price to buy at. Will be ignored for market orders.

        hidden : bool
            Whether or not to use the hidden order type.

        flags : list
            A list of integers for the different flags. Will be added together
            into a unique integer.

        Returns
        -------
        dict
            A dict containing the order detials. Used in new_order and for
            creating multiorders.
        """
        flags = flags or []
        client_order_id = wss_utils.UtcNow()
        order_op = {
            'cid': client_order_id,
            'type': order_type,
            'symbol': wss_utils.order_pair(pair),
            'amount': amount,
            'price': price,
            'hidden': hidden,
            "flags": sum(flags),
        }
        if price_trailing:
            order_op['price_trailing'] = price_trailing

        if price_aux_limit:
            order_op['price_aux_limit'] = price_aux_limit

        if price_oco_stop:
            order_op['price_oco_stop'] = price_oco_stop

        if tif:
            order_op['tif'] = tif

        return order_op

    def new_order(self, order_type, pair, amount, price, price_trailing=None,
                  price_aux_limit=None, price_oco_stop=None, hidden=0,
                  flags=None, tif=None):
        """
        Create new order.

        Parameters
        ----------
        order_type : str
            Order type. Must be one of: "LIMIT", "STOP", "MARKET",
            "TRAILING STOP", "FOK", "STOP LIMIT" or equivelent with "MARKET"
            prepended to it.

        pair : str
            The currency pair to be traded. e.g. BTCUSD

        amount :  decimal string
            The amount to be traided.

        price : decimal string
            The price to buy at. Will be ignored for market orders.

        price_trailing : decimal string
        	The trailing price

        price_aux_limit : decimal string
            Auxiliary Limit price (for STOP LIMIT)

        price_oco_stop : decimal string
            OCO stop price

        hidden : bool
            Whether or not to use the hidden order type.

        flags : list
            A list of integers for the different flags. Will be added together
            into a unique integer.

        tif : datetime string

        Returns
        -------
        int
            Order client id (cid). The CID is also a mts date stamp of when the
            order was created.
        """
        operation = self.new_order_op(
            order_type=order_type,
            pair=pair,
            amount=amount,
            price=price,
            price_trailing=price_trailing,
            price_aux_limit=price_aux_limit,
            price_oco_stop=price_oco_stop,
            hidden=hidden,
            flags=flags,
            tif=tif
        )
        data = [
            0,
            wss_utils.get_notification_code('order new'),
            None,
            operation
        ]
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)
        return operation["cid"]

    def multi_order(self, operations):
        """Multi order operation.

        Parameters
        ----------
        operations : list
            a list of operations. Read more here:
            https://bitfinex.readme.io/v2/reference#ws-input-order-multi-op
            Hint. you can use the self.new_order_op() for easy order
            operation creation.
        """
        data = [
            0,
            wss_utils.get_notification_code('order multi-op'),
            None,
            operations
        ]
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)
        return [order[1].get("cid", None) for order in operations]

    def cancel_order(self, order_id):
        """Cancel order

        Parameters
        ----------
        order_id : int, str
            Order id created by Bitfinex
        """
        data = [
            0,
            wss_utils.get_notification_code('order cancel'),
            None,
            {
                # docs: http://bit.ly/2BVqwW6
                'id': order_id
            }
        ]
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)

    def cancel_order_cid(self, order_cid, order_date):
        """Cancel order using the client id and the date of the cid. Both are
        returned from the new_order command from this library.

        Parameters
        ----------
        order_cid : str
            cid string. e.g. "1234154"

        order_date : str
            Iso formated order date. e.g. "2012-01-23"
        """
        data = [
            0,
            wss_utils.get_notification_code('order cancel'),
            None,
            {
                # docs: http://bit.ly/2BVqwW6
                'cid': order_cid,
                'cid_date': order_date
            }
        ]
        payload = json.dumps(data, ensure_ascii = False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)


    def update_order(self, **order_settings):
        """Update order using the order id

        Parameters
        ----------
        id : int64
            Order ID

        gid : int32
            Group Order ID

        price : decimal string
            Price

        amount : decimal string
            Amount

        delta : decimal string
            Change of amount

        price_aux_limit : decimal string
            Auxiliary limit price

        price_trailing : decimal string
            Trailing price delta

        tif : datetime string
            Time-In-Force: datetime for automatic order cancellation (ie. 2020-01-01 10:45:23)
        """
        data = [
            0,
            wss_utils.get_notification_code('order update'),
            None,
            order_settings
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)


    def calc(self, *calculations):
        """This message will be used by clients to trigger specific calculations,
                so we don't end up in calculating data that is not usually needed.
           You can request calculations to the websocket server that sends you the same message,
                with the required fields.
           Possible prefixes:
                margin_sym_SYMBOL (e.g. margin_sym_tBTCUSD)
                funding_sym_SYMBOL
                position_SYMBOL
                wallet_WALLET-TYPE_CURRENCY
           NOTE
                Calculations are on demand, so no more streaming of unnecessary data.
                Websocket server allows up to 30 calculations per batch.
                If the client sends too many concurrent requests (or tries to spam) requests,
                it will receive an error and potentially a disconnection.
                The Websocket server performs a maximum of 8 calculations per second per client.

            Parameters
            ----------
            *calculations : str
                margin_sym_SYMBOL
                funding_sym_SYMBOL
                position_SYMBOL
                wallet_WALLET-TYPE_CURRENCY
        """

        data = [
            0,
            'calc',
            None,
            calculations
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)



#!/usr/bin/python
#-*- coding: utf-8 -*-

from __future__ import absolute_import
import requests
import json
from json.decoder import JSONDecodeError
import base64
import hmac
import hashlib
import time

PROTOCOL = "https"
HOST = "api.bitfinex.com"
VERSION = "v1"

PATH_SYMBOLS = "symbols"
PATH_TICKER = "pubticker/%s"
PATH_TODAY = "today/%s"
PATH_STATS = "stats/%s"
PATH_LENDBOOK = "lendbook/%s"
PATH_ORDERBOOK = "book/%s"

# HTTP request timeout in seconds
TIMEOUT = 5.0


class BitfinexException(Exception):
    pass


class Client:
    """
    Client for the bitfinex.com API.

    See https://www.bitfinex.com/pages/api for API documentation.
    """

    def __init__(self, key=None, secret=None, nonce_multiplier=1.0):
        assert type(nonce_multiplier) == float, "nonce_multiplier must be decimal"
        self.URL = "%s://%s/%s" % (PROTOCOL, HOST, VERSION)
        self.BASE_URL = "%s://%s/" % (PROTOCOL, HOST)
        self.KEY = key
        self.SECRET = secret
        self.nonce_multiplier = nonce_multiplier

    def server(self):
        return u"{0:s}://{1:s}/{2:s}".format(PROTOCOL, HOST, VERSION)

    def url_for(self, path, path_arg=None, parameters=None):

        # build the basic url
        url = "%s/%s" % (self.server(), path)

        # If there is a path_arh, interpolate it into the URL.
        # In this case the path that was provided will need to have string
        # interpolation characters in it, such as PATH_TICKER
        if path_arg:
            url = url % (path_arg)

        # Append any parameters to the URL.
        if parameters:
            url = "%s?%s" % (url, self._build_parameters(parameters))

        return url

    def _nonce(self):
        """Returns a nonce used in authentication.
        Nonce must be an increasing number, if the API key has been used
        earlier or other frameworks that have used higher numbers you might
        need to increase the nonce_multiplier."""
        return str(float(time.time()) * self.nonce_multiplier)

    def _sign_payload(self, payload):
        j = json.dumps(payload)
        data = base64.standard_b64encode(j.encode('utf8'))

        h = hmac.new(self.SECRET.encode('utf8'), data, hashlib.sha384)
        signature = h.hexdigest()
        return {
            "X-BFX-APIKEY": self.KEY,
            "X-BFX-SIGNATURE": signature,
            "X-BFX-PAYLOAD": data
        }

    def _get(self, url):
        response = requests.get(url, timeout=TIMEOUT)
        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    def _post(self, endoint, payload, verify=False):
        url = self.url_for(path=endoint)
        signed_payload = self._sign_payload(payload)
        response = requests.post(url, headers=signed_payload, verify=True)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 400:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    def _build_parameters(self, parameters):
        # sort the keys so we can test easily in Python 3.3 (dicts are not
        # ordered)
        keys = list(parameters.keys())
        keys.sort()

        return '&'.join(["%s=%s" % (k, parameters[k]) for k in keys])

    def place_order(self, amount, price, side, ord_type, symbol='btcusd', exchange='bitfinex'):
        """
        Submit a new order.
        :param amount:
        :param price:
        :param side:
        :param ord_type:
        :param symbol:
        :param exchange:
        :return:
        """
        payload = {

            "request": "/v1/order/new",
            "nonce": self._nonce(),
            "symbol": symbol,
            "amount": amount,
            "price": price,
            "exchange": exchange,
            "side": side,
            "type": ord_type

        }

        response = self._post("/order/new", payload=payload, verify=True)
        return response

    def place_multiple_orders(self, orders):
        """
        Submit a new multiple order.
        """

        payload = {

           "request" : "/v1/order/new/multi",
           "nonce"   : self._nonce(),
           "orders"  : orders

        }
        response = self._post("/order/new/multi", payload=payload, verify=True)
        return response

    def delete_order(self, order_id):
        """
        Cancel an order.
        :param order_id:
        :return:
        """
        payload = {
            "request": "/v1/order/cancel",
            "nonce": self._nonce(),
            "order_id": order_id
        }

        response = self._post("/order/cancel", payload=payload, verify=True)
        return response

    def delete_all_orders(self):
        """
        Cancel all orders.

        :return:
        """
        payload = {
            "request": "/v1/order/cancel/all",
            "nonce": self._nonce(),
        }

        response = self._post("/order/cancel/all", payload=payload, verify=True)
        return response

    def status_order(self, order_id):
        """
        Get the status of an order. Is it active? Was it cancelled? To what extent has it been executed? etc.
        :param order_id:
        :return:
        """
        payload = {
            "request": "/v1/order/status",
            "nonce": self._nonce(),
            "order_id": order_id
        }

        response = self._post("/order/status", payload=payload, verify=True)
        return response

    def active_orders(self):
        """
        Fetch active orders
        """
        payload = {
            "request": "/v1/orders",
            "nonce": self._nonce()
        }
        response = self._post("orders", payload=payload, verify=True)
        return response

    def active_positions(self):
        """
        Fetch active Positions
        """
        payload = {
            "request": "/v1/positions",
            "nonce": self._nonce()
        }
        response = self._post("positions", payload=payload, verify=True)
        return response

    def claim_position(self, position_id):
        """
        Claim a position.
        :param position_id:
        :return:
        """
        payload = {
            "request": "/v1/position/claim",
            "nonce": self._nonce(),
            "position_id": position_id
        }
        response = self._post("position/claim", payload=payload, verify=True)
        return response

    def close_position(self, position_id):
        """
        Closes the selected position with a market order.
        :param position_id:
        :return:
        """
        payload = {
            "request": "/v1/position/close",
            "nonce": self._nonce(),
            "position_id": position_id
        }
        response = self._post("position/close", payload=payload, verify=True)
        return response


    def past_trades(self, timestamp='0.0', symbol='btcusd'):
        """
        Fetch past trades
        :param timestamp:
        :param symbol:
        :return:
        """
        payload = {
            "request": "/v1/mytrades",
            "nonce": self._nonce(),
            "symbol": symbol,
            "timestamp": timestamp
        }

        response = self._post("/mytrades", payload=payload, verify=True)
        return response

    def place_offer(self, currency, amount, rate, period, direction):
        """
        :param currency:
        :param amount:
        :param anual rate:
        :param period:
        :param direction:
        :return:
        """
        payload = {
            "request": "/v1/offer/new",
            "nonce": self._nonce(),
            "currency": currency,
            "amount": amount,
            "rate": rate,
            "period": period,
            "direction": direction
        }

        response = self._post("/offer/new", payload=payload, verify=True)
        return response

    def cancel_offer(self, offer_id):
        """

        :param offer_id:
        :return:
        """
        payload = {
            "request": "/v1/offer/cancel",
            "nonce": self._nonce(),
            "offer_id": offer_id
        }

        response = self._post("/offer/cancel", payload=payload, verify=True)
        return response

    def status_offer(self, offer_id):
        """

        :param offer_id:
        :return:
        """
        payload = {
            "request": "/v1/offer/status",
            "nonce": self._nonce(),
            "offer_id": offer_id
        }

        response = self._post("/offer/status", payload=payload, verify=True)
        return response


    def offers_history(self):
        """

        :param offer_id:
        :return:
        """
        payload = {
            "request": "/v1/offers/hist",
            "nonce": self._nonce()
        }
        response = self._post("/offers/hist", payload=payload, verify=True)
        return response

    def active_offers(self):
        """
        Fetch active_offers
        :return:
        """
        payload = {
            "request": "/v1/offers",
            "nonce": self._nonce()
        }

        response = self._post("/offers", payload=payload, verify=True)
        return response

    def balances(self):
        """
        Fetch balances

        :return:
        """
        payload = {
            "request": "/v1/balances",
            "nonce": self._nonce()
        }
        response = self._post("/balances", payload=payload, verify=True)
        return response

    def history(self, currency, since=0, until=9999999999, limit=500, wallet='exchange'):
        """
        View you balance ledger entries
        :param currency: currency to look for
        :param since: Optional. Return only the history after this timestamp.
        :param until: Optional. Return only the history before this timestamp.
        :param limit: Optional. Limit the number of entries to return. Default is 500.
        :param wallet: Optional. Return only entries that took place in this wallet. Accepted inputs are: “trading”,
        “exchange”, “deposit”.
        """
        payload = {
            "request": "/v1/history",
            "nonce": self._nonce(),
            "currency": currency,
            "since": since,
            "until": until,
            "limit": limit,
            "wallet": wallet
        }
        response = self._post("/history", payload=payload, verify=True)
        return response


    def movements(self, currency, start=0, end=9999999999, limit=10000, method='bitcoin'):
        """
        View you balance ledger entries
        :param currency: currency to look for
        :param since: Optional. Return only the history after this timestamp.
        :param until: Optional. Return only the history before this timestamp.
        :param limit: Optional. Limit the number of entries to return. Default is 500.
        :param method: Optional. Return only entries that used this method. Accepted inputs are: "bitcoin",
        "litecoin", "wire".
        """
        payload = {
            "request": "/v1/history/movements",
            "nonce": self._nonce(),
            "currency": currency,
            "since": start,
            "until": end,
            "limit": limit,
            "method": method
        }
        response = self._post("/history/movements", payload=payload, verify=True)
        return response

    def symbols(self):
        """
        GET /symbols

        curl https://api.bitfinex.com/v1/symbols
        ['btcusd','ltcusd','ltcbtc']
        """
        return self._get(self.url_for(PATH_SYMBOLS))


    def symbols_details(self):
        """
        GET /symbols_details

        curl https://api.bitfinex.com/v1/symbols_details
            [{
              "pair":"btcusd",
              "price_precision":5,
              "initial_margin":"30.0",
              "minimum_margin":"15.0",
              "maximum_order_size":"2000.0",
              "minimum_order_size":"0.01",
              "expiration":"NA"
            }]
        """
        return self._get(self.url_for("symbols_details"))


    def ticker(self, symbol):
        """
        GET /ticker/:symbol

        curl https://api.bitfinex.com/v1/pubticker/btcusd
        {
            'ask': '562.9999',
            'timestamp': '1395552290.70933607',
            'bid': '562.25',
            'last_price': u'562.25',
            'volume': '7842.11542563',
            'mid': u'562.62495'}
        """
        return self._get(self.url_for(PATH_TICKER, (symbol)))


    def today(self, symbol):
        """
        GET /today/:symbol

        curl "https://api.bitfinex.com/v1/today/btcusd"
        {"low":"550.09","high":"572.2398","volume":"7305.33119836"}
        """
        return self._get(self.url_for(PATH_TODAY, (symbol)))


    def stats(self, symbol):
        """
        curl https://api.bitfinex.com/v1/stats/btcusd
        [
            {"period":1,"volume":"7410.27250155"},
            {"period":7,"volume":"52251.37118006"},
            {"period":30,"volume":"464505.07753251"}
        ]
        """
        data = self._get(self.url_for(PATH_STATS, (symbol)))

        for period in data:

            for key, value in period.items():
                if key == 'period':
                    new_value = int(value)
                elif key == 'volume':
                    new_value = value

                period[key] = new_value

        return data


    def lendbook(self, currency, parameters=None):
        """
        curl "https://api.bitfinex.com/v1/lendbook/btc"

        {"bids":[{"rate":"5.475","amount":"15.03894663","period":30,"timestamp":"1395112149.0","frr":"No"},{"rate":"2.409","amount":"14.5121868","period":7,"timestamp":"1395497599.0","frr":"No"}],"asks":[{"rate":"6.351","amount":"15.5180735","period":5,"timestamp":"1395549996.0","frr":"No"},{"rate":"6.3588","amount":"626.94808249","period":30,"timestamp":"1395400654.0","frr":"Yes"}]}

        Optional parameters

        limit_bids (int): Optional. Limit the number of bids (loan demands) returned. May be 0 in which case the array of bids is empty. Default is 50.
        limit_asks (int): Optional. Limit the number of asks (loan offers) returned. May be 0 in which case the array of asks is empty. Default is 50.
        """
        data = self._get(self.url_for(PATH_LENDBOOK, path_arg=currency, parameters=parameters))

        for lend_type in data.keys():

            for lend in data[lend_type]:

                for key, value in lend.items():
                    if key in ['rate', 'amount', 'timestamp']:
                        new_value = value
                    elif key == 'period':
                        new_value = int(value)
                    elif key == 'frr':
                        new_value = value == 'Yes'

                    lend[key] = new_value

        return data


    def order_book(self, symbol, parameters=None):
        """
        curl "https://api.bitfinex.com/v1/book/btcusd"

        {"bids":[{"price":"561.1101","amount":"0.985","timestamp":"1395557729.0"}],"asks":[{"price":"562.9999","amount":"0.985","timestamp":"1395557711.0"}]}

        The 'bids' and 'asks' arrays will have multiple bid and ask dicts.

        Optional parameters

        limit_bids (int): Optional. Limit the number of bids returned. May be 0 in which case the array of bids is empty. Default is 50.
        limit_asks (int): Optional. Limit the number of asks returned. May be 0 in which case the array of asks is empty. Default is 50.

        eg.
        curl "https://api.bitfinex.com/v1/book/btcusd?limit_bids=1&limit_asks=0"
        {"bids":[{"price":"561.1101","amount":"0.985","timestamp":"1395557729.0"}],"asks":[]}

        """
        data = self._get(self.url_for(PATH_ORDERBOOK, path_arg=symbol, parameters=parameters))

        for type_ in data.keys():
            for list_ in data[type_]:
                for key, value in list_.items():
                    list_[key] = value

        return data

class TradeClient(Client):
    """Added for backward compatibility"""
    pass



#!/usr/bin/python
#-*- coding: utf-8 -*-
"""Bitfinex Rest API V2 implementation"""
# pylint: disable=R0904

from __future__ import absolute_import
import json
from json.decoder import JSONDecodeError
import hmac
import hashlib
import time
import requests



PROTOCOL = "https"
HOST = "api.bitfinex.com"
VERSION = "v2"


# HTTP request timeout in seconds
TIMEOUT = 5.0


class BitfinexException(Exception):
    """
    Exception handler
    """
    pass


class Client:
    """
    : Client for the bitfinex.com API REST V2.
    : See these links for documentation :
    :     https://bitfinex.readme.io/v2/docs
    :     https://bitfinex.readme.io/v2/reference
    """

    def __init__(self, key=None, secret=None, nonce_multiplier=1.0):
        """
        Object initialisation takes 2 mandatory arguments key and secret and a optional one
        nonce_multiplier
        """

        assert isinstance(nonce_multiplier, float), "nonce_multiplier must be decimal"
        self.base_url = "%s://%s/" % (PROTOCOL, HOST)
        self.key = key
        self.secret = secret
        self.nonce_multiplier = nonce_multiplier

    def _nonce(self):
        """Returns a nonce used in authentication.
        Nonce must be an increasing number, if the API key has been used
        earlier or other frameworks that have used higher numbers you might
        need to increase the nonce_multiplier."""
        return str(float(time.time()) * self.nonce_multiplier)

    def _headers(self, path, nonce, body):
        """
        create signed headers
        """
        signature = "/api/" + path + nonce + body
        hmc = hmac.new(self.secret.encode('utf8'), signature.encode('utf8'), hashlib.sha384)
        signature = hmc.hexdigest()

        return {
            "bfx-nonce": nonce,
            "bfx-apikey": self.key,
            "bfx-signature": signature,
            "content-type": "application/json"
        }

    def _post(self, path, payload, verify=False):
        """
        Send post request to bitfinex
        """
        nonce = self._nonce()
        headers = self._headers(path, nonce, payload)
        response = requests.post(self.base_url + path, headers=headers, data=payload, verify=verify)

        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)



    def _get(self, path, **params):
        """
        Send get request to bitfinex
        """
        url = self.base_url + path
        response = requests.get(url, timeout=TIMEOUT, params=params)
        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)




    # REST PUBLIC ENDPOINTS
    def platform_status(self):
        """
        Get the current status of the platform. Maintenance periods last for just few minutes and
        might be necessary from time to time during upgrades of core components of our
        infrastructure. Even if rare it is important to have a way to notify users. For a real-time
        notification we suggest to use websockets and listen to events 20060/20061
        https://bitfinex.readme.io/v2/reference#rest-public-platform-status
        Response details : 	int	1=operative, 0=maintenance
        """
        path = "v2/platform/status"
        response = self._get(path)
        return response



    def tickers(self, symbol_list):
        """
        The ticker is a high level overview of the state of the market. It shows you the current
        best bid and ask, as well as the last trade price.It also includes information such as daily
        volume and how much the price has moved over the last day.
        https://bitfinex.readme.io/v2/reference#rest-public-tickers
        """
        assert isinstance(symbol_list, list), "symbol_list must be of type list"
        assert symbol_list, "symbol_list must have at least one symbol"
        path = "v2/tickers?symbols={}".format(",".join(symbol_list))
        response = self._get(path)
        return response

    def ticker(self, symbol):
        """
        The ticker is a high level overview of the state of the market.It shows you the current best
        bid and ask, as well as the last trade price.It also includes information such as daily
        volume and how much the price has moved over the last day.
        https://bitfinex.readme.io/v2/reference#rest-public-ticker
        """
        path = "v2/ticker/"+str(symbol)
        response = self._get(path)
        return response



    def trades(self, symbol):
        """
        Trades endpoint includes all the pertinent details of the trade, such as price,
        size and time.
        https://bitfinex.readme.io/v2/reference#rest-public-trades
        """
        path = "v2/trades/"+symbol+"/hist"
        response = self._get(path)
        return response


    def books(self, symbol, precision="P0"):
        """
        The Order Books channel allow you to keep track of the state of the Bitfinex order book.
        It is provided on a price aggregated basis, with customizable precision.
        https://bitfinex.readme.io/v2/reference#rest-public-books
        Params :
            symbol :The symbol you want information about. You can find the list of valid symbols
                    by calling the /symbols endpoint.
            precision : Level of price aggregation (P0, P1, P2, P3, R0)
        """
        path = f"v2/book/{symbol}/{precision}"
        response = self._get(path)
        return response


    def stats(self, **kwargs):
        """
        Various statistics about the requested pair.
        https://bitfinex.readme.io/v2/reference#rest-public-stats
        Params :
            Key     : Allowed values: "funding.size", "credits.size", "credits.size.sym", "pos.size"
            Size    : Available values: '1m'
            Symbol  : The symbol you want information about.
            Symbol2 : The symbol you want information about.
            Side    : Available values: "long", "short"
            Section : Available values: "last", "hist"
            sort    : 0 , 1
        """
        key_values = ['funding.size', 'credits.size', 'credits.size.sym', 'pos.size']
        if kwargs['key'] not in key_values:
            raise ValueError("Key must have one of the following values : "+" ".join(key_values))


        common_stats_url = "v2/stats1/{key}:{size}:{symbol}".format(
            key=kwargs['key'],
            size=kwargs['size'],
            symbol=kwargs['symbol']
        )

        if kwargs['key'] == 'pos.size':
            custom_stats_url = ":{side}/{section}?sort={sort}".format(
                side=kwargs['side'],
                section=kwargs['section'],
                sort=str(kwargs['sort'])
            )

        if kwargs['key'] in ['funding.size', 'credits.size']:
            custom_stats_url = "/{section}?sort={sort}".format(
                section=kwargs['section'],
                sort=str(kwargs['sort'])
            )

        if kwargs['key'] == 'credits.size.sym':
            custom_stats_url = ":{symbol2}/{section}?sort={sort}".format(
                symbol2=kwargs['symbol2'],
                section=kwargs['section'],
                sort=str(kwargs['sort'])
            )

        path = "".join([common_stats_url, custom_stats_url])

        response = self._get(path)
        return response

    def candles(self, *args, **kwargs):
        """
        Provides a way to access charting candle info
        https://bitfinex.readme.io/v2/reference#rest-public-candles
        """
        margs = list(args)
        section = margs.pop()
        path = "v2/candles/trade"
        for arg in margs:
            path = path + ":" + arg
        path += "/"+section
        response = self._get(path, **kwargs)
        return response



    #REST CALCULATION ENDPOINTS
    def market_average_price(self, **kwargs):
        """
        Calculate the average execution rate for Trading or Margin funding.
        https://bitfinex.readme.io/v2/reference#rest-calc-market-average-price
        """
        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/calc/trade/avg"
        response = self._post(path, raw_body, verify=True)
        return response


    def foreign_exchange_rate(self, **kwargs):
        """
        https://bitfinex.readme.io/v2/reference#foreign-exchange-rate
        """
        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/calc/fx"
        response = self._post(path, raw_body, verify=True)
        return response




    #REST AUTHENTICATED ENDPOINTS
    def wallets_balance(self):
        """
        Get account wallets
        https://bitfinex.readme.io/v2/reference#rest-auth-wallets
        """

        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/wallets"
        response = self._post(path, raw_body, verify=True)
        return response


    def active_orders(self, trade_pair=""):
        """
        Fetch active orders using rest api v2
        https://bitfinex.readme.io/v2/reference#rest-auth-orders
        """

        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/orders/"+trade_pair
        response = self._post(path, raw_body, verify=True)
        return response


    def orders_history(self, trade_pair, **kwargs):
        """
        Returns the most recent closed or canceled orders up to circa two weeks ago
        https://bitfinex.readme.io/v2/reference#orders-history
        """

        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/auth/r/orders/"+trade_pair+"/hist"
        response = self._post(path, raw_body, verify=True)
        return response

    def order_trades(self, trade_pair, order_id):
        """
        Get Trades generated by an Order
        https://bitfinex.readme.io/v2/reference#rest-auth-order-trades
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/order/"+trade_pair+":"+str(order_id)+"/trades"
        response = self._post(path, raw_body, verify=True)
        return response


    def trades_history(self, trade_pair, **kwargs):
        """
        Returns list of trades
        https://api.bitfinex.com/v:version/auth/r/trades/:Symbol/hist
        """

        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/auth/r/trades/"+trade_pair+"/hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def active_positions(self):
        """
        Returns list of active Positions
        https://bitfinex.readme.io/v2/reference#rest-auth-positions
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/positions"
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_offers(self, symbol=""):
        """
        Get active funding offers.
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-offers
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/offers/"+symbol
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_offers_history(self, symbol="", **kwargs):
        """
        Get past inactive funding offers. Limited to last 3 days.
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-offers-hist
        """
        body = kwargs
        raw_body = json.dumps(body)
        addsymbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/offers/"+addsymbol+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_loans(self, symbol=""):
        """
        Funds not used in active positions
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-loans
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/loans/"+symbol
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_loans_history(self, symbol="", **kwargs):
        """
        Inactive funds not used in positions. Limited to last 3 days.
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-loans-hist
        """
        body = kwargs
        raw_body = json.dumps(body)
        addsymbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/loans/"+addsymbol+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_credits(self, symbol=""):
        """
        Funds used in active positions
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-credits
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/credits/"+symbol
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_credits_history(self, symbol="", **kwargs):
        """
        Inactive funds used in positions. Limited to last 3 days.
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-credits-hist
        """
        body = kwargs
        raw_body = json.dumps(body)
        addsymbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/credits/"+addsymbol+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_trades(self, symbol="", **kwargs):
        """
        Get funding trades
        https://bitfinex.readme.io/v2/reference#rest-auth-funding-trades-hist
        """
        body = kwargs
        raw_body = json.dumps(body)
        addsymbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/trades/"+addsymbol+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def margin_info(self, tradepair="base"):
        """
        Get account margin info
        https://bitfinex.readme.io/v2/reference#rest-auth-info-margin
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/info/margin/"+tradepair
        response = self._post(path, raw_body, verify=True)
        return response


    def funding_info(self, tradepair):
        """
        Get account funding info
        https://bitfinex.readme.io/v2/reference#rest-auth-info-funding
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/info/funding/"+tradepair
        response = self._post(path, raw_body, verify=True)
        return response


    def movements(self, currency=""):
        """
        View your past deposits/withdrawals.
        https://bitfinex.readme.io/v2/reference#movements
        """
        body = {}
        raw_body = json.dumps(body)
        addcurrency = "{}/".format(currency.upper()) if currency else ""
        path = "v2/auth/r/movements/"+addcurrency+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def performance(self, period="1D"):
        """
        Get account historical daily performance (work in progress on Bitfinex side)
        Work in progress
        This endpoint is still under active development so you might experience unexpected behavior
        from it.
        https://bitfinex.readme.io/v2/reference#rest-auth-performance

        Currently not working : bitfinex.rest.restv2.BitfinexException:
        (500, 'Internal Server Error', ['error', 10020, 'method: invalid'])
        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/stats/perf:"+period+"/hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def alert_list(self):
        """
        List of active alerts
        https://api.bitfinex.com/v:version/auth/r/alerts
        """
        body = {'type':'price'}
        raw_body = json.dumps(body)
        path = "v2/auth/r/alerts"
        response = self._post(path, raw_body, verify=True)
        return response


    def alert_set(self, alert_type, symbol, price):
        """
        Sets up a price alert at the given value
        https://bitfinex.readme.io/v2/reference#rest-auth-alert-set
        """
        body = {
            'type'   : alert_type,
            'symbol' : symbol,
            'price'  : price
        }

        raw_body = json.dumps(body)
        path = "v2/auth/w/alert/set"
        response = self._post(path, raw_body, verify=True)
        return response


    def alert_delete(self, symbol, price):
        """
        https://bitfinex.readme.io/v2/reference#rest-auth-alert-delete
        Bitfinex always returns [True] no matter if the request deleted an alert or not
        """
        body = {}

        raw_body = json.dumps(body)
        path = "v2/auth/w/alert/price:"+symbol+":"+str(price)+"/del"
        response = self._post(path, raw_body, verify=True)
        return response


    def calc_available_balance(self, symbol, direction, rate, order_type):
        """
        Calculate available balance for order/offer
        https://bitfinex.readme.io/v2/reference#rest-auth-calc-bal-avail
        example : calc_available_balance("tIOTUSD","1","1.13","EXCHANGE")
        QUERY PARAMS :
            symbol : symbol (string)
            dir    : direction of the order/offer
                     (orders: > 0 buy, < 0 sell | offers: > 0 sell, < 0 buy) (integer)
            rate   : Rate of the order/offer (string)
            type   : Type of the order/offer EXCHANGE or MARGIN (string)
        """

        body = {
            'symbol' : symbol,
            'dir'    : direction,
            'rate'   : rate,
            'type'   : order_type
        }

        raw_body = json.dumps(body)
        path = "v2/auth/calc/order/avail"
        response = self._post(path, raw_body, verify=True)
        return response


    def ledgers(self, currency=""):
        """
        View your past ledger entries.
        https://bitfinex.readme.io/v2/reference#ledgers
        """
        body = {}
        raw_body = json.dumps(body)
        addcurrency = "{}/".format(currency.upper()) if currency else ""
        path = "v2/auth/r/ledgers/"+addcurrency+"hist"
        response = self._post(path, raw_body, verify=True)
        return response


    def user_settings_read(self, pkey):
        """
        Read user settings
        https://bitfinex.readme.io/v2/reference#user-settings-read
        """
        body = {'keys':['api:'+str(pkey)]}
        raw_body = json.dumps(body)
        path = "v2/auth/r/settings"
        response = self._post(path, raw_body, verify=True)
        return response


    def user_settings_write(self, pkey):
        """
        Write user settings
        https://bitfinex.readme.io/v2/reference#user-settings-write
        """
        raise NotImplementedError


    def user_settings_delete(self, pkey):
        """
        Delete user settings
        https://bitfinex.readme.io/v2/reference#user-settings-delete
        """
        raise NotImplementedError



"""Module for websocket utilities"""
import datetime

def UtcNow():
    now = datetime.datetime.utcnow()
    return int(float(now.strftime("%s.%f"))*10000000)

def order_pair(pair):
    pair = 't' + pair if not pair.startswith('t') else pair
    return pair

ERROR_CODES = {
    10000:	"Unknown error",
    10001:	"Generic error",
    10008:	"Concurrency error",
    10020:	"Request parameters error",
    10050:	"Configuration setup failed",
    10100:	"Failed authentication",
    10111:	"Error in authentication request payload",
    10112:	"Error in authentication request signature",
    10113:	"Error in authentication request encryption",
    10114:	"Error in authentication request nonce",
    10200:	"Error in un-authentication request",
    10300:	"Failed channel subscription",
    10301:	"Failed channel subscription: already subscribed",
    10400:	"Failed channel un-subscription: channel not found",
    11000:	"Not ready, try again later",
    20051:	"Websocket server stopping... please reconnect later",
    20060:	"Websocket server resyncing... please reconnect later",
    20061:	"Websocket server resync complete. please reconnect",
    5000:	"Info message",
}

# Abbreviation Glossary
# https://bitfinex.readme.io/v2/docs/abbreviations-glossary#section-abbreviation-glossary
NOTIFICATION_CODES = {
    "bu":	"balance update",
    "ps":	"position snapshot",
    "pn":	"new position",
    "pu":	"position update",
    "pc":	"position close",
    "ws":	"wallet snapshot",
    "wu":	"wallet update",
    "os":	"order snapshot",
    "on":	"order new",
    "ou":	"order update",
    "oc":	"order cancel",
    "ox_multi": "order multi-op",
    "oc-req":	"order cancel request",
    "oc_multi-req":	"multiple orders cancel request",
    "te":	"trade executed",
    "tu":	"trade execution update",
    "fte":	"funding trade execution",
    "ftu":	"funding trade update",
    "hos":	"historical order snapshot",
    "mis":	"margin information snapshot",
    "miu":	"margin information update",
    "n":	"notification",
    "fos":	"funding offer snapshot",
    "fon":	"funding offer new",
    "fou":	"funding offer update",
    "foc":	"funding offer cancel",
    "hfos":	"historical funding offer snapshot",
    "fcs":	"funding credits snapshot",
    "fcn":	"funding credits new",
    "fcu":	"funding credits update",
    "fcc":	"funding credits close",
    "hfcs":	"historical funding credits snapshot",
    "fls":	"funding loan snapshot",
    "fln":	"funding loan new",
    "flu":	"funding loan update",
    "flc":	"funding loan close",
    "hfls":	"historical funding loan snapshot",
    "hfts":	"historical funding trade snapshot",
    "uac":	"user custom price alert",
}

def get_notification_code(description):
    index = list(NOTIFICATION_CODES.values()).index(description)
    return list(NOTIFICATION_CODES.keys())[index]

ORDER_TYPES = [
    "MARKET",
    "EXCHANGE MARKET",
    "LIMIT",
    "EXCHANGE LIMIT",
    "STOP",
    "EXCHANGE STOP",
    "TRAILING STOP",
    "EXCHANGE TRAILING STOP",
    "FOK",
    "EXCHANGE FOK",
    "STOP LIMIT",
    "EXCHANGE STOP LIMIT",
]



from setuptools import setup, find_packages

VERSION = '0.8.1'

# Runtime dependencies. See requirements.txt for development dependencies.
DEPENDENCIES = [
    'requests',
    'httpretty',
    "twisted",
    "autobahn",
    "pyopenssl",
    "service_identity",
]

setup(
    name='bitfinex-v2',
    version=VERSION,
    description='Python client for the Bitfinex API',
    author='Ole Henrik Skogstrøm',
    author_email='henrik@amplify.no',
    url='https://github.com/ohenrik/bitfinex',
    license='MIT',
    packages=find_packages(),
    install_requires=DEPENDENCIES,
    # download_url='https://github.com/ohenrik/bitfinex/tarball/%s' % version,
    keywords=['bitcoin', 'btc'],
    classifiers=[],
    zip_safe=True
)
